-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local LibTSMTypes = select(2, ...).LibTSMTypes
local Group = LibTSMTypes:Init("Group")
local GroupItemMap = LibTSMTypes:IncludeClassType("GroupItemMap")
local GroupItemDB = LibTSMTypes:IncludeClassType("GroupItemDB")
local Path = LibTSMTypes:Include("Group.Path")
local TempTable = LibTSMTypes:From("LibTSMUtil"):Include("BaseType.TempTable")
local String = LibTSMTypes:From("LibTSMUtil"):Include("Lua.String")
local Table = LibTSMTypes:From("LibTSMUtil"):Include("Lua.Table")
local Log = LibTSMTypes:From("LibTSMUtil"):Include("Util.Log")
local private = {
	operationInfo = nil,
	operationExistsFunc = nil,
	itemMap = nil, ---@type GroupItemMap
	itemDB = nil, ---@type GroupItemDB
	groups = nil,
	items = nil,
}
local DEFAULT_OPERATION_NAME = "#Default"



-- ============================================================================
-- Module Loading
-- ============================================================================

Group:OnModuleLoad(function()
	private.itemMap = GroupItemMap.New(Group.IsItemInGroup)
	private.itemDB = GroupItemDB.New("GROUP_ITEMS")
end)



-- ============================================================================
-- Module Functions
-- ============================================================================

---Returns whether or not a group name is valid.
---@param name string The name
---@return boolean
function Group.IsValidName(name)
	return Path.IsValidName(name)
end

---Gets the root group path.
---@return GroupPathValue
function Group.GetRootPath()
	return Path.GetRoot()
end

---Gets the name of the group from a group path.
---@param groupPath GroupPathValue The group path
---@return string
function Group.GetName(groupPath)
	return Path.GetName(groupPath)
end

---Gets the parent group path or nil for the given group path.
---@param groupPath GroupPathValue The group path
---@return GroupPathValue?
function Group.GetParent(groupPath)
	return Path.GetParent(groupPath)
end

---Gets the relative path between two group paths.
---@param groupPath GroupPathValue The group path
---@param prefixGroupPath GroupPathValue The base group path component
---@return GroupPathValue
function Group.GetRelativePath(groupPath, prefixGroupPath)
	return Path.GetRelative(groupPath, prefixGroupPath)
end

---Checks whether a group path is the child of another.
---@param groupPath GroupPathValue The group path to check
---@param parentPath GroupPathValue The parent group path to compare with
---@return boolean
function Group.IsChild(groupPath, parentPath)
	return Path.IsChild(groupPath, parentPath)
end

---Gets the top level group for a group path.
---@param groupPath GroupPathValue The group path
---@return GroupPathValue
function Group.GetTopLevel(groupPath)
	return Path.GetTopLevel(groupPath)
end

---Creates a new group path by joining the components.
---@param ... string The components
---@return GroupPathValue
function Group.JoinPath(...)
	return Path.Join(...)
end

---Splits a group path into the parent group path and name.
---@param groupPath GroupPathValue The group path
---@return GroupPathValue?
---@return string
function Group.SplitPath(groupPath)
	return Path.Split(groupPath)
end

---Formats a group path for display to the user.
---@param groupPath GroupPathValue The group path
---@return string
function Group.FormatPath(groupPath)
	return Path.Format(groupPath)
end

---Converts the group path into a sortable representation.
---@param groupPath GroupPathValue The group path
---@return string
function Group.GetSortableString(groupPath)
	return Path.GetSortableString(groupPath)
end

---Gets the level of a group path (top-level groups are 1)
---@param groupPath GroupPathValue The group path
---@return number
function Group.GetLevel(groupPath)
	return Path.GetLevel(groupPath)
end

---Sets the operation info.
---@param info table<string,number> Table with the key being the operation type and the value being the max number of that operation
---@param existsFunc fun(typeName: string, operationName: string): boolean Function to check if an operation exists
function Group.SetOperationInfo(info, existsFunc)
	assert(type(info) == "table" and type(existsFunc) == "function")
	private.operationInfo = private.operationInfo or {}
	wipe(private.operationInfo)
	for typeName, maxNum in pairs(info) do
		assert(type(typeName) == "string" and type(maxNum) == "number")
		private.operationInfo[typeName] = maxNum
	end
	private.operationExistsFunc = existsFunc
end

---Loads group settings.
---@param groups table<GroupPathValue,table>
---@param items table<string,GroupPathValue>
function Group.Load(groups, items)
	private.groups = groups
	private.items = items
	private.SanitizeGroups()
	private.SanitizeItems()
	private.itemMap:Invalidate()
	private.itemDB:Load(private.items)
end

---Iterates over all groups.
---@return fun(): string, table @Iterator with fields: `groupPath`, `groupOperations`
function Group.Iterator()
	return pairs(private.groups)
end

---Iterates over operations assigned to a group.
---@param groupPath GroupPathValue The group path
---@param typeName string The type of the operation
---@return fun(): number, string @Iterator with fields: `index`, `operationName`
function Group.OperationIterator(groupPath, typeName)
	return ipairs(private.groups[groupPath][typeName])
end

---Gets the number of operations assigned to a group.
---@param groupPath GroupPathValue The group path
---@param typeName string The type of the operation
---@return number
function Group.GetNumOperations(groupPath, typeName)
	return #private.groups[groupPath][typeName]
end

---Returns whether or a specific operation is applied to the group.
---@param groupPath GroupPathValue The group path
---@param typeName string The type of the operation
---@param operationName string The name of the operation
---@return boolean
function Group.HasOperation(groupPath, typeName, operationName)
	return tContains(private.groups[groupPath][typeName], operationName) and true or false
end

---Gets the group item DB for joining.
---@return DatabaseTable
function Group.GetItemDBForJoin()
	return private.itemDB:GetForJoin()
end

---Creates an item DB query.
---@param groupPathFilter GroupPathValue The group path to filter on
---@param includeSubGroups boolean Whether or not to include subgroups
---@return DatabaseQuery
function Group.CreateItemsQuery(groupPathFilter, includeSubGroups)
	return private.itemDB:CreateQuery(groupPathFilter, includeSubGroups)
end

---Iterates over items in a group.
---@param groupPathFilter GroupPathValue The group path to filter on
---@param includeSubGroups boolean Whether or not to include subgroups
---@return fun(): number, string, string @Iterator with fields: `index`, `itemString`, `groupPath`
function Group.ItemIterator(groupPathFilter, includeSubGroups)
	return private.itemDB:Iterator(groupPathFilter, includeSubGroups)
end

---Iterates over items in a group by their base item string.
---@param baseItemString any
---@return fun(): number, string, string @Iterator with fields: `index`, `itemString`, `groupPath`
function Group.ItemByBaseItemStringIterator(baseItemString)
	return private.itemDB:IteratorByBaseItem(baseItemString)
end

---Gets the number of items in a group.
---@param groupPathFilter GroupPathValue The group path to filter on
---@return number
function Group.GetNumItems(groupPathFilter)
	return private.itemDB:CreateQuery(groupPathFilter):CountAndRelease()
end

---Sets whether or not item DB query updates are paused.
---@param paused boolean Whether or not query updates are paused
function Group.SetItemDBQueriesPaused(paused)
	private.itemDB:SetQueryUpdatesPaused(paused)
end

---Translates an itemString based on what's grouped.
---@param itemString string The item string
---@return string
function Group.TranslateItemString(itemString)
	return private.itemMap:Translate(itemString)
end

---Returns whether or not a group exists.
---@param groupPath GroupPathValue The group path
---@return boolean
function Group.Exists(groupPath)
	return private.groups[groupPath] and true or false
end

---Creates a new group.
---@param groupPath GroupPathValue The group path
function Group.Create(groupPath)
	local createPaths = TempTable.Acquire()
	tinsert(createPaths, groupPath)
	local parentPath = Path.GetParent(groupPath)
	assert(parentPath)
	-- Traverse up the tree and get all the parent groups that don't yet exists
	while not Path.IsRoot(parentPath) and not Group.Exists(parentPath) do
		tinsert(createPaths, parentPath)
		parentPath = Path.GetParent(parentPath)
	end
	-- Create from the top of the tree down
	while #createPaths > 0 do
		private.CreateGroup(tremove(createPaths))
	end
end

---Moves a group to a new path.
---@param groupPath GroupPathValue The existing group path
---@param newGroupPath GroupPathValue The new group path
function Group.Move(groupPath, newGroupPath)
	assert(Group.Exists(groupPath) and not Group.Exists(newGroupPath))
	assert(not Path.IsRoot(groupPath))
	local newParentPath = Path.GetParent(newGroupPath)
	assert(newParentPath and Group.Exists(newParentPath), "Parent of target is invalid")

	-- Get a list of group path changes for this group and all its subgroups
	local changes = TempTable.Acquire()
	local gsubEscapedNewGroupPath = gsub(newGroupPath, "%%", "%%%%")
	for path in Group.Iterator() do
		if path == groupPath or Path.IsChild(path, groupPath) then
			changes[path] = gsub(path, "^"..String.Escape(groupPath), gsubEscapedNewGroupPath)
		end
	end

	Group.SetItemDBQueriesPaused(true)
	local itemsMoved = TempTable.Acquire()
	for oldPath, newPath in pairs(changes) do
		-- Move the group
		assert(Group.Exists(oldPath) and not Group.Exists(newPath))
		private.groups[newPath] = private.groups[oldPath]
		private.groups[oldPath] = nil

		-- Move the items
		wipe(itemsMoved)
		private.itemDB:Move(oldPath, newPath, itemsMoved)
		for _, itemString in ipairs(itemsMoved) do
			assert(private.items[itemString])
			private.items[itemString] = newPath
		end
	end
	TempTable.Release(itemsMoved)
	TempTable.Release(changes)
	Group.SetItemDBQueriesPaused(false)

	for typeName in pairs(private.operationInfo) do
		if not Group.HasOperationOverride(newGroupPath, typeName) then
			Group.InheritParentOperations(newGroupPath, typeName)
			Group.UpdateChildGroupOperations(groupPath, typeName)
		end
	end
end

---Deletes a group.
---@param groupPath GroupPathValue The group path
function Group.Delete(groupPath)
	assert(not Path.IsRoot(groupPath) and private.groups[groupPath])

	-- Delete this group and all subgroups
	for path in Group.Iterator() do
		if path == groupPath or Path.IsChild(path, groupPath) then
			-- Delete this group
			private.groups[path] = nil
		end
	end

	-- Remove all items from our DB
	Group.SetItemDBQueriesPaused(true)
	local query = Group.CreateItemsQuery(groupPath, true)
		:Select("itemString")
	local updateMapItems = TempTable.Acquire()
	for _, itemString in query:Iterator() do
		assert(private.items[itemString])
		private.items[itemString] = nil
		private.itemDB:SetGroup(itemString, nil)
		updateMapItems[itemString] = true
	end
	query:Release()
	private.itemMap:ValuesChanged(updateMapItems)
	TempTable.Release(updateMapItems)
	Group.SetItemDBQueriesPaused(false)
end

---Returns whether or not an item is in a group.
---@param itemString string The item string
---@return boolean
function Group.IsItemInGroup(itemString)
	return private.itemDB:HasItem(itemString)
end

---Gets the group path for an item.
---@param itemString string The item string
---@return GroupPathValue
function Group.GetPathByItem(itemString)
	local groupPath = private.itemDB:GetGroup(Group.TranslateItemString(itemString))
	assert(Group.Exists(groupPath))
	return groupPath
end

---Sets the group an item is in.
---@param itemString string The item string
---@param groupPath GroupPathValue The group path
function Group.SetItemGroup(itemString, groupPath)
	private.SetItemGroup(itemString, groupPath)
end

---Sets the group a list of items are in.
---@param items string[] The item strings
---@param groupPath GroupPathValue The group path
function Group.SetItemsGroup(items, groupPath)
	Group.SetItemDBQueriesPaused(true)
	for _, itemString in ipairs(items) do
		private.SetItemGroup(itemString, groupPath)
	end
	Group.SetItemDBQueriesPaused(false)
end

---Imports a set of items into a group and subgroups and returns the number of items added.
---@param groupName GroupPathValue The top-level group name to import into
---@param items table<string,GroupPathValue> The items and their relative group paths
---@param moveExistingItems boolean Whether or not to move items which are already in a group
---@param groupOperations table<GroupPathValue,table> The operations to import to the group
---@return number
function Group.ImportItemsAndOperations(groupName, items, moveExistingItems, groupOperations)
	assert(not Path.IsRoot(groupName))
	local numItems = 0
	for itemString, relGroupPath in pairs(items) do
		local groupPath = relGroupPath == Path.GetRoot() and groupName or Path.Join(groupName, relGroupPath)
		assert(Group.Exists(groupPath))
		if not Group.IsItemInGroup(itemString) then
			private.items[itemString] = groupPath
			numItems = numItems + 1
		elseif moveExistingItems then
			private.items[itemString] = groupPath
			numItems = numItems + 1
		end
	end
	for relGroupPath, moduleOperations in pairs(groupOperations) do
		local groupPath = relGroupPath == "" and groupName or Path.Join(groupName, relGroupPath)
		for moduleName, operations in pairs(moduleOperations) do
			if operations.override then
				private.groups[groupPath][moduleName] = operations
				Group.UpdateChildGroupOperations(groupPath, moduleName)
			end
		end
	end
	Group.SetItemDBQueriesPaused(true)
	private.itemDB:Load(private.items)
	private.itemMap:Invalidate()
	Group.SetItemDBQueriesPaused(false)
	return numItems
end

---Inherits operations of a given type from the parent group.
---@param groupPath GroupPathValue The group path
---@param typeName string The type of the operation
function Group.InheritParentOperations(groupPath, typeName)
	local parentGroupPath = Path.GetParent(groupPath)
	local groupOperations = private.groups[groupPath][typeName]
	local override = groupOperations.override
	wipe(groupOperations)
	groupOperations.override = override
	for _, operationName in ipairs(private.groups[parentGroupPath][typeName]) do
		tinsert(groupOperations, operationName)
	end
end

---Returns whether or not a group is overriding operations of a given type.
---@param groupPath GroupPathValue The group path
---@param typeName string The type of the operation
---@return boolean
function Group.HasOperationOverride(groupPath, typeName)
	return private.groups[groupPath][typeName].override and true or false
end

---Updates the operations of all children of a group.
---@param groupPath GroupPathValue The group path
---@param typeName string The type of the operation
function Group.UpdateChildGroupOperations(groupPath, typeName)
	for childGroupPath in Group.Iterator() do
		if Path.IsChild(childGroupPath, groupPath) and not Group.HasOperationOverride(childGroupPath, typeName) then
			Group.InheritParentOperations(childGroupPath, typeName)
		end
	end
end

---Handles an operation being renamed.
---@param typeName string The type of the operation
---@param oldName string The old operation name
---@param newName string The new operation name
function Group.HandleOperationRenamed(typeName, oldName, newName)
	-- Just blindly rename in all groups - no need to check for override
	for _, info in Group.Iterator() do
		for i = 1, #info[typeName] do
			if info[typeName][i] == oldName then
				info[typeName][i] = newName
			end
		end
	end
end

---Handles an operation being deleted.
---@param typeName string The type of the operation
---@param operationName string The name of the operation
function Group.HandleOperationDeleted(typeName, operationName)
	-- Just blindly remove from all groups - no need to check for override
	for _, groupOperations in Group.Iterator() do
		Table.RemoveByValue(groupOperations[typeName], operationName)
	end
end

---Swaps the position of operations assigned to a group.
---@param groupPath GroupPathValue The group path
---@param typeName string The type of the operation
---@param index1 number The first index to swap
---@param index2 number The second index to swap
function Group.SwapOperation(groupPath, typeName, index1, index2)
	local groupOperations = private.groups[groupPath][typeName]
	groupOperations[index1], groupOperations[index2] = groupOperations[index2], groupOperations[index1]
	Group.UpdateChildGroupOperations(groupPath, typeName)
end

---Removes an operation from a group.
---@param groupPath GroupPathValue The group path
---@param typeName string The type of the operation
---@param operation number|string Either the index or name of the operation to remove
function Group.RemoveOperation(groupPath, typeName, operation)
	local groupOperations = private.groups[groupPath][typeName]
	assert(groupOperations.override)
	if type(operation) == "number" then
		assert(groupOperations[operation])
		tremove(groupOperations, operation)
	elseif type(operation) == "string" then
		assert(Table.RemoveByValue(groupOperations, operation) > 0)
	else
		error("Invalid operation: "..tostring(operation))
	end
	Group.UpdateChildGroupOperations(groupPath, typeName)
end

---Appens an operation to the group.
---@param groupPath GroupPathValue The group path
---@param typeName string The type of the operation
---@param operation number|string Either the index or name of the operation to remove
function Group.AppendOperation(groupPath, typeName, operationName)
	assert(private.operationExistsFunc(typeName, operationName) and Group.GetNumOperations(groupPath, typeName) < private.operationInfo[typeName])
	local groupOperations = private.groups[groupPath][typeName]
	assert(groupOperations.override)
	tinsert(groupOperations, operationName)
	Group.UpdateChildGroupOperations(groupPath, typeName)
end

---Sets whether or not a group is overriding operations.
---@param groupPath GroupPathValue The group path
---@param typeName string The type of the operation
---@param override boolean Whether or not to override
function Group.SetOperationOverride(groupPath, typeName, override)
	if override == Group.HasOperationOverride(groupPath, typeName) then
		return
	end
	assert(not Path.IsRoot(groupPath))
	local groupOperations = private.groups[groupPath][typeName]
	if override then
		wipe(groupOperations)
		groupOperations.override = true
		Group.UpdateChildGroupOperations(groupPath, typeName)
	else
		groupOperations.override = nil
		Group.InheritParentOperations(groupPath, typeName)
		Group.UpdateChildGroupOperations(groupPath, typeName)
	end
end

---Exports a group.
---@param exportGroupPath GroupPathValue The group path to export
---@param includedSubGroups table<GroupPathValue,boolean> Relative subgroup paths to include in the export
---@param resultGroups table<GroupPathValue,boolean> Table to store result groups in
---@param resultItems table<string,GroupPathValue> Table to store result items in
function Group.Export(exportGroupPath, includedSubGroups, resultGroups, resultItems)
	for groupPath in Group.Iterator() do
		local relGroupPath = nil
		if Group.IsChild(groupPath, exportGroupPath) then
			relGroupPath = Group.GetRelativePath(groupPath, exportGroupPath)
			if not includedSubGroups[relGroupPath] then
				relGroupPath = nil
			end
		elseif groupPath == exportGroupPath then
			relGroupPath = Group.GetRootPath()
		end
		if relGroupPath then
			resultGroups[relGroupPath] = true
			for _, itemString in Group.ItemIterator(groupPath) do
				resultItems[itemString] = relGroupPath
			end
		end
	end
end



-- ============================================================================
-- Private Helper Functions
-- ============================================================================

function private.SanitizeGroups()
	for groupPath, groupInfo in Group.Iterator() do
		-- Clear ignoreRandomEnchants and ignoreItemVariations
		groupInfo.ignoreItemVariations = nil
		groupInfo.ignoreRandomEnchants = nil

		-- Remove operations which no longer exist
		for typeName, operations in pairs(groupInfo) do
			for i = #operations, 1, -1 do
				if not private.operationExistsFunc(typeName, operations[i]) then
					Log.Err("Removing invalid operation from group (%s): %s, %s", groupPath, typeName, tostring(operations[i]))
					tremove(operations, i)
				end
			end
		end

		if type(groupPath) == "string" and Path.IsValid(groupPath) then
			-- Check the contents of groupInfo
			for typeName in pairs(private.operationInfo) do
				groupInfo[typeName] = groupInfo[typeName] or {}
				if Path.IsRoot(groupPath) then
					-- Root group should have override flag set
					groupInfo[typeName].override = true
				end
			end
			for key in pairs(groupInfo) do
				if private.operationInfo[key] then
					-- This is a set of module operations
					local operations = groupInfo[key]
					while #operations > private.operationInfo[key] do
						-- Remove extra operations
						tremove(operations)
					end
					for key2 in pairs(operations) do
						if key2 == "override" then
							-- Ensure the override field is either true or nil
							operations.override = operations.override and true or nil
						elseif type(key2) ~= "number" or key2 <= 0 or key2 > #operations then
							-- This is an invalid key
							Log.Err("Removing invalid operations key (%s, %s): %s", groupPath, key, tostring(key2))
							operations[key2] = nil
						end
					end
					for i = #operations, 1, -1 do
						-- Clean up invalid operations
						if type(operations[i]) ~= "string" or operations[i] == "" then
							tremove(operations, i)
						end
					end
				else
					-- Invalid key
					Log.Err("Removing invalid groupInfo key (%s): %s", groupPath, tostring(key))
					groupInfo[key] = nil
				end
			end
		else
			-- Remove invalid group paths
			Log.Err("Removing invalid group path: %s", tostring(groupPath))
			private.groups[groupPath] = nil
		end
	end

	-- Create the root group if it doesn't exist
	if not private.groups[Path.GetRoot()] then
		-- Set the override flag for all top-level groups
		for groupPath, groupOperations in Group.Iterator() do
			if Path.IsTopLevel(groupPath) then
				for typeName in pairs(private.operationInfo) do
					groupOperations[typeName].override = true
				end
			end
		end
		-- Create the root group with default operations applied
		local groupOperations = {}
		private.groups[Path.GetRoot()] = groupOperations
		for typeName in pairs(private.operationInfo) do
			groupOperations[typeName] = { DEFAULT_OPERATION_NAME, override = true }
		end
	end

	-- Remove any groups without a parent
	for _, groupPath in private.OrderedGroupPathIterator() do
		local parentPath = Path.GetParent(groupPath)
		if groupPath ~= Path.GetRoot() and not private.groups[parentPath] then
			-- The parent group doesn't exist, so remove this group
			Log.Err("Removing group with non-existent parent: %s", tostring(groupPath))
			private.groups[groupPath] = nil
		end
	end

	-- Clean up operation inheritence
	for _, groupPath in private.OrderedGroupPathIterator() do
		if groupPath ~= Path.GetRoot() then
			for typeName in pairs(private.operationInfo) do
				if not Group.HasOperationOverride(groupPath, typeName) then
					Group.InheritParentOperations(groupPath, typeName)
				end
			end
		end
	end
end

function private.SanitizeItems()
	-- Fix up any invalid items
	local newPaths = TempTable.Acquire()
	for itemString, groupPath in pairs(private.items) do
		local newItemString = GroupItemMap.SanitizeItemString(itemString)
		if not newItemString then
			-- This itemstring is invalid
			Log.Err("Itemstring (%s) is invalid", tostring(itemString))
			private.items[itemString] = nil
		elseif Path.IsRoot(groupPath) or not Group.Exists(groupPath) then
			-- This group doesn't exist
			Log.Err("Group (%s) doesn't exist, so removing item (%s)", groupPath, itemString)
			private.items[itemString] = nil
		elseif newItemString ~= itemString then
			-- Remove this invalid itemstring from this group
			Log.Err("Itemstring changed (%s -> %s), so removing it from group (%s)", itemString, newItemString, groupPath)
			private.items[itemString] = nil
			-- Add this new item to this group if it's not already in one
			if not private.items[newItemString] then
				newPaths[newItemString] = groupPath
				Log.Err("Adding to group instead (%s)", groupPath)
			end
		end
	end
	for itemString, groupPath in pairs(newPaths) do
		private.items[itemString] = groupPath
	end
	TempTable.Release(newPaths)
end

function private.OrderedGroupPathIterator()
	local result = TempTable.Acquire()
	for groupPath in Group.Iterator() do
		tinsert(result, groupPath)
	end
	Path.SortPaths(result)
	return TempTable.Iterator(result)
end

function private.SetItemGroup(itemString, groupPath)
	assert(not groupPath or (not Path.IsRoot(groupPath) and Group.Exists(groupPath)))
	local hadItem = private.itemDB:HasItem(itemString)
	private.itemDB:SetGroup(itemString, groupPath)
	private.items[itemString] = groupPath
	if not hadItem or not groupPath then
		private.itemMap:ValueChanged(itemString)
	end
end

function private.CreateGroup(groupPath)
	assert(not Group.Exists(groupPath))
	local parentPath = Path.GetParent(groupPath)
	assert(parentPath and (Path.IsRoot(parentPath) or Group.Exists(parentPath)))
	private.groups[groupPath] = {}
	for typeName in pairs(private.operationInfo) do
		private.groups[groupPath][typeName] = {}
		Group.InheritParentOperations(groupPath, typeName)
	end
end
