-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local LibTSMWoW = select(2, ...).LibTSMWoW
local ItemClass = LibTSMWoW:Init("Util.ItemClass")
local ClientInfo = LibTSMWoW:Include("Util.ClientInfo")
local private = {
	classes = {},
	subClasses = {},
	classLookup = {},
	classIdLookup = {},
	inventorySlotIdLookup = {},
}



-- ============================================================================
-- Data
-- ============================================================================

local RETAIL_ITEM_CLASS_IDS = {
	Enum.ItemClass.Weapon,
	Enum.ItemClass.Armor,
	Enum.ItemClass.Container,
	Enum.ItemClass.Gem,
	Enum.ItemClass.ItemEnhancement,
	Enum.ItemClass.Consumable,
	Enum.ItemClass.Glyph,
	Enum.ItemClass.Tradegoods,
	Enum.ItemClass.Recipe,
	Enum.ItemClass.Profession,
	Enum.ItemClass.Battlepet,
	Enum.ItemClass.Questitem,
	Enum.ItemClass.Miscellaneous,
}
local WRATH_ITEM_CLASS_IDS = {
	Enum.ItemClass.Weapon,
	Enum.ItemClass.Armor,
	Enum.ItemClass.Container,
	Enum.ItemClass.Consumable,
	Enum.ItemClass.Glyph,
	Enum.ItemClass.Tradegoods,
	Enum.ItemClass.Projectile,
	Enum.ItemClass.Quiver,
	Enum.ItemClass.Recipe,
	Enum.ItemClass.Gem,
	Enum.ItemClass.Miscellaneous,
	Enum.ItemClass.Questitem,
}
local VANILLA_ITEM_CLASS_IDS = {
	Enum.ItemClass.Weapon,
	Enum.ItemClass.Armor,
	Enum.ItemClass.Container,
	Enum.ItemClass.Consumable,
	Enum.ItemClass.Tradegoods,
	Enum.ItemClass.Projectile,
	Enum.ItemClass.Quiver,
	Enum.ItemClass.Recipe,
	Enum.ItemClass.Reagent,
	Enum.ItemClass.Miscellaneous,
}



-- ============================================================================
-- Module Loading
-- ============================================================================

ItemClass:OnModuleLoad(function()
	local data = nil
	if LibTSMWoW.IsRetail() then
		data = RETAIL_ITEM_CLASS_IDS
	elseif LibTSMWoW.IsWrathClassic() then
		data = WRATH_ITEM_CLASS_IDS
	elseif LibTSMWoW.IsVanillaClassic() then
		data = VANILLA_ITEM_CLASS_IDS
	else
		error("Unknown game version")
	end

	for _, classId in ipairs(data) do
		local class = GetItemClassInfo(classId)
		if class then
			private.classIdLookup[strlower(class)] = classId
			private.classLookup[class] = {}
			private.classLookup[class]._index = classId
			local subClasses = nil
			if ClientInfo.HasFeature(ClientInfo.FEATURES.C_AUCTION_HOUSE) then
				subClasses = C_AuctionHouse.GetAuctionItemSubClasses(classId)
			else
				subClasses = {GetAuctionItemSubClasses(classId)}
			end
			for _, subClassId in pairs(subClasses) do
				local subClassName = GetItemSubClassInfo(classId, subClassId)
				if not strfind(subClassName, "(OBSOLETE)") then
					private.classLookup[class][subClassName] = subClassId
				end
			end
		end
	end

	for class, subClasses in pairs(private.classLookup) do
		tinsert(private.classes, class)
		private.subClasses[class] = {}
		for subClass in pairs(subClasses) do
			if subClass ~= "_index" then
				tinsert(private.subClasses[class], subClass)
			end
		end
		sort(private.subClasses[class], function(a, b) return private.classLookup[class][a] < private.classLookup[class][b] end)
	end
	sort(private.classes, function(a, b) return private.classIdLookup[strlower(a)] < private.classIdLookup[strlower(b)] end)

	for _, id in pairs(Enum.InventoryType) do
		local invType = GetItemInventorySlotInfo(id)
		if invType then
			private.inventorySlotIdLookup[strlower(invType)] = id
		end
	end
end)



-- ============================================================================
-- Module Functions
-- ============================================================================

---Gets the pet class ID.
---@return number
function ItemClass.GetPetClassId()
	return Enum.ItemClass.Battlepet
end

---Gets all item class names.
---@return string[]
function ItemClass.GetClasses()
	return private.classes
end

---Gets all item sub class names.
---@param class string The class name
---@return string[]
function ItemClass.GetSubClasses(class)
	return private.subClasses[class]
end

---Gets the class ID.
---@param classStr string The class name
---@return number
function ItemClass.GetClassIdFromClassString(classStr)
	return private.classIdLookup[strlower(classStr)]
end

---Gets the sub class ID.
---@param subClass string The sub class name
---@param classId number The class ID
---@return number
function ItemClass.GetSubClassIdFromSubClassString(subClass, classId)
	if not classId then return end
	local class = GetItemClassInfo(classId)
	if not private.classLookup[class] then return end
	for str, index in pairs(private.classLookup[class]) do
		if strlower(str) == strlower(subClass) then
			return index
		end
	end
end

---Gets the inventory slot ID by name.
---@param slot string The name
---@return number
function ItemClass.GetInventorySlotIdFromInventorySlotString(slot)
	return private.inventorySlotIdLookup[strlower(slot)]
end
